package de.uniulm.sai;

import java.awt.*;
import java.util.Hashtable;

/**
* TableLayout - A layout manager for Swing that allows to manage
*		components in a table like structure.
* Copyright (C) 2003-2006	Johannes Mayer
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*
* Known Problems:
* - There's no error handling for the positioning of GUI element
*   right now.
*
* @author  Johannes Mayer
* @version 0.9.1, 2006-10-31
*/
public class TableLayout implements LayoutManager2 {

    /**
    * Horizontal alignment (left, center, right, fill)
    */
    public static class AlignX {
	public static final AlignX LEFT = new AlignX();
	public static final AlignX CENTER = new AlignX();
	public static final AlignX RIGHT = new AlignX();
	public static final AlignX FILLX = new AlignX();
	private AlignX() {
	}
	// Apply the settings to gbc
	private void applySettingsTo(GridBagConstraints gbc) {
	    if (this == FILLX) {
		if (gbc.fill == GridBagConstraints.NONE)
		    gbc.fill = GridBagConstraints.HORIZONTAL;
		else if (gbc.fill == GridBagConstraints.VERTICAL)
		    gbc.fill = GridBagConstraints.BOTH;
	    }
	    else {
		if (gbc.fill == GridBagConstraints.HORIZONTAL)
		    gbc.fill = GridBagConstraints.NONE;
		else if (gbc.fill == GridBagConstraints.BOTH)
		    gbc.fill = GridBagConstraints.VERTICAL;
	    }

	    if (this == LEFT) {
		if (gbc.anchor == GridBagConstraints.NORTH)
		    gbc.anchor = GridBagConstraints.NORTHWEST;
		else if (gbc.anchor == GridBagConstraints.NORTHEAST)
		    gbc.anchor = GridBagConstraints.NORTHWEST;
		else if (gbc.anchor == GridBagConstraints.CENTER)
		    gbc.anchor = GridBagConstraints.WEST;
		else if (gbc.anchor == GridBagConstraints.EAST)
		    gbc.anchor = GridBagConstraints.WEST;
		else if (gbc.anchor == GridBagConstraints.SOUTH)
		    gbc.anchor = GridBagConstraints.SOUTHWEST;
		else if (gbc.anchor == GridBagConstraints.SOUTHEAST)
		    gbc.anchor = GridBagConstraints.SOUTHWEST;
	    }
	    else if (this == CENTER || this == FILLX) {
		if (gbc.anchor == GridBagConstraints.NORTHWEST)
		    gbc.anchor = GridBagConstraints.NORTH;
		else if (gbc.anchor == GridBagConstraints.NORTHEAST)
		    gbc.anchor = GridBagConstraints.NORTH;
		else if (gbc.anchor == GridBagConstraints.WEST)
		    gbc.anchor = GridBagConstraints.CENTER;
		else if (gbc.anchor == GridBagConstraints.EAST)
		    gbc.anchor = GridBagConstraints.CENTER;
		else if (gbc.anchor == GridBagConstraints.SOUTHWEST)
		    gbc.anchor = GridBagConstraints.SOUTH;
		else if (gbc.anchor == GridBagConstraints.SOUTHEAST)
		    gbc.anchor = GridBagConstraints.SOUTH;
	    }
	    else if (this == RIGHT) {
		if (gbc.anchor == GridBagConstraints.NORTHWEST)
		    gbc.anchor = GridBagConstraints.NORTHEAST;
		else if (gbc.anchor == GridBagConstraints.NORTH)
		    gbc.anchor = GridBagConstraints.NORTHEAST;
		else if (gbc.anchor == GridBagConstraints.WEST)
		    gbc.anchor = GridBagConstraints.EAST;
		else if (gbc.anchor == GridBagConstraints.CENTER)
		    gbc.anchor = GridBagConstraints.EAST;
		else if (gbc.anchor == GridBagConstraints.SOUTHWEST)
		    gbc.anchor = GridBagConstraints.SOUTHEAST;
		else if (gbc.anchor == GridBagConstraints.SOUTH)
		    gbc.anchor = GridBagConstraints.SOUTHEAST;
	    }
	}
    }

    /**
    * Vertical alignment (top, center, bottom, fill)
    */
    public static class AlignY {
	public static final AlignY TOP = new AlignY();
	public static final AlignY CENTER = new AlignY();
	public static final AlignY BOTTOM = new AlignY();
	public static final AlignY FILLY = new AlignY();
	private AlignY() {
	}
	// Apply the settings to gbc
	private void applySettingsTo(GridBagConstraints gbc) {
	    if (this == FILLY) {
		if (gbc.fill == GridBagConstraints.NONE)
		    gbc.fill = GridBagConstraints.VERTICAL;
		else if (gbc.fill == GridBagConstraints.HORIZONTAL)
		    gbc.fill = GridBagConstraints.BOTH;
	    }
	    else {
		if (gbc.fill == GridBagConstraints.VERTICAL)
		    gbc.fill = GridBagConstraints.NONE;
		else if (gbc.fill == GridBagConstraints.BOTH)
		    gbc.fill = GridBagConstraints.HORIZONTAL;
	    }

	    if (this == TOP) {
		if (gbc.anchor == GridBagConstraints.WEST)
		    gbc.anchor = GridBagConstraints.NORTHWEST;
		else if (gbc.anchor == GridBagConstraints.CENTER)
		    gbc.anchor = GridBagConstraints.NORTH;
		else if (gbc.anchor == GridBagConstraints.EAST)
		    gbc.anchor = GridBagConstraints.NORTHEAST;
		else if (gbc.anchor == GridBagConstraints.SOUTHWEST)
		    gbc.anchor = GridBagConstraints.NORTHWEST;
		else if (gbc.anchor == GridBagConstraints.SOUTH)
		    gbc.anchor = GridBagConstraints.NORTH;
		else if (gbc.anchor == GridBagConstraints.SOUTHEAST)
		    gbc.anchor = GridBagConstraints.NORTHEAST;
	    }
	    else if (this == CENTER || this == FILLY) {
		if (gbc.anchor == GridBagConstraints.NORTHWEST)
		    gbc.anchor = GridBagConstraints.WEST;
		else if (gbc.anchor == GridBagConstraints.NORTH)
		    gbc.anchor = GridBagConstraints.CENTER;
		else if (gbc.anchor == GridBagConstraints.NORTHEAST)
		    gbc.anchor = GridBagConstraints.EAST;
		else if (gbc.anchor == GridBagConstraints.SOUTHWEST)
		    gbc.anchor = GridBagConstraints.WEST;
		else if (gbc.anchor == GridBagConstraints.SOUTH)
		    gbc.anchor = GridBagConstraints.CENTER;
		else if (gbc.anchor == GridBagConstraints.SOUTHEAST)
		    gbc.anchor = GridBagConstraints.EAST;
	    }
	    else if (this == BOTTOM) {
		if (gbc.anchor == GridBagConstraints.NORTHWEST)
		    gbc.anchor = GridBagConstraints.SOUTHWEST;
		else if (gbc.anchor == GridBagConstraints.NORTH)
		    gbc.anchor = GridBagConstraints.SOUTH;
		else if (gbc.anchor == GridBagConstraints.NORTHEAST)
		    gbc.anchor = GridBagConstraints.SOUTHEAST;
		else if (gbc.anchor == GridBagConstraints.WEST)
		    gbc.anchor = GridBagConstraints.SOUTHWEST;
		else if (gbc.anchor == GridBagConstraints.CENTER)
		    gbc.anchor = GridBagConstraints.SOUTH;
		else if (gbc.anchor == GridBagConstraints.EAST)
		    gbc.anchor = GridBagConstraints.SOUTHEAST;
	    }
	}
    }
    
    /**
    * Constraints for the positioning and formatting
    * of GUI objects within a TableLayout
    */
    public static class Constraints implements Cloneable {
    	// position within the table
	private Integer x = null, y = null, dx = null, dy = null;
	// alignment within the cell
	private AlignX ax = null;
	private AlignY ay = null;
	// distances
	private Integer insetTop = null;
	private Integer insetBottom = null;
	private Integer insetLeft = null;
	private Integer insetRight = null;
	private Integer ipadx = null, ipady = null;
	
	public Constraints() {
	    this.dx = new Integer(1);
	    this.dy = new Integer(1);
	}
	public Constraints(int x, int y, int dx, int dy) {
	    if (x < 0 || y < 0 || dx < 1 || dy < 1)
		throw new IllegalArgumentException();
	    this.x = new Integer(x);
	    this.y = new Integer(y);
	    this.dx = new Integer(dx);
	    this.dy = new Integer(dy);
	}
	public Constraints(int x, int y) {
	    this(x, y, 1, 1);
	}
	
	public Object clone() {
	    Constraints c = null;
	    try {
		c = (Constraints) super.clone();
	    }
	    catch (CloneNotSupportedException e) {
	    }
	    return c;
	}
	public Constraints pos(int x, int y) {
	    if (x < 0 || y < 0)
		throw new IllegalArgumentException();
	    Constraints c = (Constraints) this.clone();
	    c.x = new Integer(x);
	    c.y = new Integer(y);
	    return c;
	}
	public Constraints size(int dx, int dy) {
	    if (dx < 1 || dy < 1)
		throw new IllegalArgumentException();
	    Constraints c = (Constraints) this.clone();
	    c.dx = new Integer(dx);
	    c.dy = new Integer(dy);
	    return c;
	}
	public Constraints alignX(AlignX ax) {
	    Constraints c = (Constraints) this.clone();
	    c.ax = ax;
	    return c;
	}
	public Constraints alignY(AlignY ay) {
	    Constraints c = (Constraints) this.clone();
	    c.ay = ay;
	    return c;
	}
	public Constraints aligns(AlignX ax, AlignY ay) {
	    Constraints c = (Constraints) this.clone();
	    c.ax = ax;
	    c.ay = ay;
	    return c;
	}
	public Constraints insetsX(int left, int right) {
	    if (left < 0 || right < 0)
		throw new IllegalArgumentException();
	    Constraints c = (Constraints) this.clone();
	    c.insetLeft = new Integer(left);
	    c.insetRight = new Integer(right);
	    return c;
	}
	public Constraints insetsY(int top, int bottom) {
	    if (top < 0 || bottom < 0)
		throw new IllegalArgumentException();
	    Constraints c = (Constraints) this.clone();
	    c.insetTop = new Integer(top);
	    c.insetBottom = new Integer(bottom);
	    return c;
	}
	public Constraints insets(int top, int left, int bottom, int right) {
	    if (top < 0 || left < 0 || bottom < 0 || right < 0)
		throw new IllegalArgumentException();
	    Constraints c = (Constraints) this.clone();
	    c.insetTop = new Integer(top);
	    c.insetLeft = new Integer(left);
	    c.insetBottom = new Integer(bottom);
	    c.insetRight = new Integer(right);
	    return c;
	}
	public Constraints ipadX(int ipadx) {
	    if (ipadx < 0)
		throw new IllegalArgumentException();
	    Constraints c = (Constraints) this.clone();
	    c.ipadx = new Integer(ipadx);
	    return c;
	}
	public Constraints ipadY(int ipady) {
	    if (ipady < 0)
		throw new IllegalArgumentException();
	    Constraints c = (Constraints) this.clone();
	    c.ipady = new Integer(ipady);
	    return c;
	}
	public Constraints ipad(int ipadx, int ipady) {
	    if (ipadx < 0 || ipady < 0)
		throw new IllegalArgumentException();
	    Constraints c = (Constraints) this.clone();
	    c.ipadx = new Integer(ipadx);
	    c.ipady = new Integer(ipady);
	    return c;
	}
	// Apply the settings to gbc
	private void applySettingsTo(GridBagConstraints gbc) {
	    if (x != null)
	        gbc.gridx = x.intValue();
	    if (y != null)
	        gbc.gridy = y.intValue();
	    if (dx != null)
	        gbc.gridwidth = dx.intValue();
	    if (dy != null)
	        gbc.gridheight = dy.intValue();
	    if (ax != null)
	        ax.applySettingsTo(gbc);
	    if (ay != null)
	        ay.applySettingsTo(gbc);
	    if (insetTop != null)
	        gbc.insets.top = insetTop.intValue();
	    if (insetBottom != null)
	        gbc.insets.bottom = insetBottom.intValue();
	    if (insetLeft != null)
	        gbc.insets.left = insetLeft.intValue();
	    if (insetRight != null)
	        gbc.insets.right = insetRight.intValue();
	    if (ipadx != null)
	        gbc.ipadx = ipadx.intValue();
	    if (ipady != null)
	        gbc.ipady = ipady.intValue();
	}
    }

    // internal GridBagLayout object that does the layout job
    private GridBagLayout l = new GridBagLayout();

    // global constraints for all cells of the table
    private GridBagConstraints c = new GridBagConstraints();
    
    // constraints for columns and rows
    private Hashtable cols = new Hashtable();
    private Hashtable rows = new Hashtable();
    
    // weights for columns and rows
    private Hashtable colStretch = new Hashtable();
    private Hashtable rowStretch = new Hashtable();
    
    // next cell to be filled
    private int nextx = 0, nexty = 0;
    
    // number of columns of this TableLayout
    private int columns = Integer.MAX_VALUE;

    public TableLayout(int columns) {
	if (columns <= 0)
	    throw new IllegalArgumentException("columns must not be negative");
	this.columns = columns;
	
    	// set global defaults
	setAlignX(AlignX.CENTER);
	setAlignY(AlignY.CENTER);
	setInsets(0, 0, 0, 0);
	setIpad(0, 0);
    }

    // settings for all cells

    public void setAlignX(AlignX ax) {
	ax.applySettingsTo(c);
    }
    public void setAlignY(AlignY ay) {
	ay.applySettingsTo(c);
    }
    public void setInsets(int top, int left, int bottom, int right) {
	if (top < 0 || left < 0 || bottom < 0 || right < 0)
	    throw new IllegalArgumentException("all arguments must not be negative");
	c.insets = new Insets(top, left, bottom, right);
    }
    public void setIpad(int ipadx, int ipady) {
	if (ipadx < 0 || ipady < 0)
	    throw new IllegalArgumentException();
        c.ipadx = ipadx;
        c.ipady = ipady;
    }

    // settings for all columns

    public void setColumnAlign(int c, AlignX ax) {
	if (c < 0)
	    throw new IllegalArgumentException();
	Constraints cc = (Constraints) cols.get(new Integer(c));
	if (cc == null)
	    cc = new Constraints();
	cc = cc.alignX(ax);
	cols.put(new Integer(c), cc);
    }
    public void setColumnInsets(int c, int left, int right) {
	if (c < 0 || left < 0 || right < 0)
	    throw new IllegalArgumentException();
	Constraints cc = (Constraints) cols.get(new Integer(c));
	if (cc == null)
	    cc = new Constraints();
	cc = cc.insetsX(left, right);
	cols.put(new Integer(c), cc);
    }
    public void setColumnIpad(int c, int ipadx) {
	if (c < 0 || ipadx < 0)
	    throw new IllegalArgumentException();
	Constraints cc = (Constraints) cols.get(new Integer(c));
	if (cc == null)
	    cc = new Constraints();
	cc = cc.ipadX(ipadx);
	cols.put(new Integer(c), cc);
    }
    public void setColumnStretch(int c, double s) {
	if (c < 0 || s < 0.0)
	    throw new IllegalArgumentException();
	colStretch.put(new Integer(c), new Double(s));
    }

    // settings for all rows

    public void setRowAlign(int r, AlignY ay) {
	if (r < 0)
	    throw new IllegalArgumentException();
	Constraints rc = (Constraints) rows.get(new Integer(r));
	if (rc == null)
	    rc = new Constraints();
	rc = rc.alignY(ay);
	cols.put(new Integer(r), rc);
    }
    public void setRowInsets(int r, int top, int bottom) {
	if (r < 0 || top < 0 || bottom < 0)
	    throw new IllegalArgumentException();
	Constraints rc = (Constraints) rows.get(new Integer(r));
	if (rc == null)
	    rc = new Constraints();
	rc = rc.insetsY(top, bottom);
	cols.put(new Integer(r), rc);
    }
    public void setRowIpad(int r, int ipady) {
	if (r < 0 || ipady < 0)
	    throw new IllegalArgumentException();
	Constraints rc = (Constraints) rows.get(new Integer(r));
	if (rc == null)
	    rc = new Constraints();
	rc = rc.ipadY(ipady);
	cols.put(new Integer(r), rc);
    }
    public void setRowStretch(int r, double s) {
	if (r < 0 || s < 0.0)
	    throw new IllegalArgumentException();
	rowStretch.put(new Integer(r), new Double(s));
    }

    // apply the settings for columns x, ..., x+dx to gbc
    private void applyColumnSettings(GridBagConstraints gbc, int x, int dx) {
	// add stretches
	double sum = 0.0;
	for (int i = x; i < x+dx; i++) {
	    Integer xi = new Integer(i);
	    if (colStretch.containsKey(xi)) {
		Double s = (Double) colStretch.get(xi);
		sum += s.doubleValue();
	    }
	}
        gbc.weightx = sum;

	// constraints
	Integer xi = new Integer(x);
	if (cols.containsKey(xi))
	    ((Constraints) cols.get(xi)).applySettingsTo(gbc);

    }
    // apply the settings for rows y, ..., y+dy to gbc
    private void applyRowSettings(GridBagConstraints gbc, int y, int dy) {
	// add stretches
	double sum = 0.0;
	for (int i = y; i < y+dy; i++) {
	    Integer yi = new Integer(i);
	    if (rowStretch.containsKey(yi)) {
		Double s = (Double) rowStretch.get(yi);
		sum += s.doubleValue();
	    }
	}
	gbc.weighty = sum;

	// constraints
	Integer yi = new Integer(y);
	if (rows.containsKey(yi))
	    ((Constraints) rows.get(yi)).applySettingsTo(gbc);

    }

    public void addLayoutComponent(Component comp, Object constraints) {
	if (constraints == null)
	    constraints = new Constraints(nextx, nexty);

	Constraints cs = (Constraints) constraints;

	GridBagConstraints cnew = (GridBagConstraints) c.clone();
	cnew.insets = (Insets) c.insets.clone();

	// position from the cell contraints
	if (cs.x == null)
	    cs = cs.pos(nextx, nexty);
	cnew.gridx = cs.x.intValue();
	cnew.gridy = cs.y.intValue();

	// and also width and height ...
	if (cs.dx == null)
	    cs = cs.size(1, 1);
	cnew.gridwidth = cs.dx.intValue();
	cnew.gridheight = cs.dy.intValue();

	// update the position for the next GUI element
	nextx = cnew.gridx + cnew.gridwidth;
	nexty = cnew.gridy;
	if (nextx >= columns) {
	    nextx = 0;
	    nexty += cnew.gridheight;
	}

	// Priority of settings
	// 1. cell constraints
	// 2. row constraints
	// 3. column constraints
	// 4. global constraints

	// apply column and row constrains
	applyColumnSettings(cnew, cnew.gridx, cnew.gridwidth);
	applyRowSettings(cnew, cnew.gridy, cnew.gridheight);

	// apply cell constraints
	cs.applySettingsTo(cnew);

	// delegate the call to l
	l.addLayoutComponent(comp, cnew);
    }
    
    // delegate all calls to l
    public void removeLayoutComponent(Component comp) {
	l.removeLayoutComponent(comp);
    }
    public Dimension maximumLayoutSize(Container target) {
        return l.maximumLayoutSize(target);
    }
    public Dimension preferredLayoutSize(Container parent) {
    	return l.preferredLayoutSize(parent);
    }
    public Dimension minimumLayoutSize(Container parent) {
	return l.minimumLayoutSize(parent);
    }
    public float getLayoutAlignmentX(Container target) {
        return l.getLayoutAlignmentX(target);
    }
    public float getLayoutAlignmentY(Container target) {
	return l.getLayoutAlignmentY(target);
    }
    public void layoutContainer(Container parent) {
	l.layoutContainer(parent);
    }
    public void invalidateLayout(Container target) {
	l.invalidateLayout(target);
    }
    // not supported
    public void addLayoutComponent(String name, Component comp) {
	throw new UnsupportedOperationException();
    }
}
